
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;

public class Entlehnungsverwaltung {

	private Connection conn = null;
	
	public void dbConnect() throws Exception{
		
		String driver = "jdbc:postgresql";
		String host = "bordo.dbai.tuwien.ac.at";
		String port = "5432";
		String database = "database";
		String username = "username";
		String password = "password";
		
		String dbConnectionString = driver + "://" + host + ":" + port + "/" + database;
		
		
		Class.forName("org.postgresql.Driver");
		conn = DriverManager.getConnection(dbConnectionString, username, password);
		conn.setAutoCommit(false);
	}
	
	
	/**
	 * Methode "dbOutputKunden()", die alle Entlehnkartenbesitzer (Entlehnkartennummer, Vor- und Nachname, 
	 * offene Gebhren auf der Entlehnkarte) ausgibt, die Anzahl aller bisherigen Entlehnungen sowie die 
	 * Gesamtsumme der Gebhren, die durch alle Entlehnungen dieses Kunden bisher entstanden sind. Vergessen 
	 * Sie nicht auf jene Kunden, fr die noch keine Entlehnungen erfasst sind! Verwenden Sie dabei die 
	 * Funktion "f_gebuehr", die Sie in Punkt 6.2 erstellt haben. Die Ausgabe soll als Comma-Separated List 
	 * erfolgen, d.h.: eine Bildschirm-Zeile pro Zeile der Tabelle; die Spaltenwerte werden einfach nacheinander 
	 * ausgegeben und mittels "," getrennt.
	 */
	public void dbOutputKunden() throws SQLException{
		
		Statement stmt = null;
		ResultSet rs = null;
		
		try {
			stmt = conn.createStatement();
			
			rs = stmt.executeQuery("SELECT k.nr as entlehnkartennr, " +
								   "p.vname as vorname, " +
								   "p.nname as nachname, " +
								   "k.gebuehr as offene_gebuehren, " +
								   "COUNT(e.eid) as anz_entlehnungen, " +
								   "SUM(f_gebuehr(e.eid)) as gesamt_entlehngebuehren " +
								   "FROM Entlehnkarte k " +
								   "JOIN Person p ON (k.person = p.pid) " +
								   "LEFT OUTER JOIN Entlehnung e ON (k.nr = e.entlehnkarte) " +
					               "GROUP BY k.nr, p.vname, p.nname, k.gebuehr " +
					               "ORDER BY k.nr");
			
			while (rs.next()) {
				System.out.print(rs.getString("entlehnkartennr") + ", ");
				System.out.print(rs.getString("vorname") + ", ");
				System.out.print(rs.getString("nachname") + ", ");
				System.out.print(rs.getString("offene_gebuehren") + ", ");
				System.out.print(rs.getString("anz_entlehnungen") + ", ");
				System.out.println(rs.getString("gesamt_entlehngebuehren"));
			}
			
			rs.close();
			stmt.close();
		} catch (Exception e) {
			System.out.println(e.toString());
			
			if (rs != null) {
				rs.close();
			}
			if (stmt != null) {
				stmt.close();
			}
			
		}
		
	}
	
	/*
	 * Methode "editGebuehren()", die exakt die Funktionalitaet der PL/pgSQL-Prozedur von Punkt 6.3 bereitstellt. 
	 * Beachten Sie bei der Erstellung dieser Methode folgende Punkte:
	 * - "editGebuehren()" soll die Funktion "f_gebuehr" aus Punkt 6.2 verwenden. Ansonsten duerfen keine weiteren 
	 *   Benutzer-definierten Funktionen oder Prozeduren verwendet werden. Insbesonders darf nicht einfach die Prozedur 
	 *   aus Punkt 6.3 aufgerufen werden.
	 * - Stellen Sie durch entsprechende Transaction Control Kommandos sicher, dass die Methode "editGebuehren()" entweder 
	 *   die vorgesehene Aenderung erfolgreich durchfuehrt (und festschreibt) oder die Datenbank unveraendert laesst.
	 * - Vermeiden Sie in der Java-Methode die Verwendung von "*" in SELECT-Anweisungen. Listen Sie statt dessen in jeder 
	 *   SELECT Anweisung explizit die Spalten auf, die Sie auslesen (oder veraendern) moechten, z.B.: In der Uni-DB wuerden 
	 *   Sie "SELECT MatrNr, Name, Semester FROM Studenten" statt "SELECT * FROM Studenten" schreiben. Und wenn das Semester 
	 *   in der Programmlogik keine Rolle spielt, wuerden Sie "SELECT MatrNr, Name FROM Studenten" schreiben.
	 */
	public void editGebuehren(java.sql.Date dat) throws SQLException{
		
		Statement stmt = null;
		PreparedStatement pstmt = null;
		String s;
		
		try {
			
			stmt = conn.createStatement();
			
			/* Alle Entlehnungen und dafr anfallende Gebuehren an diesem Tag herausfinden */
			ResultSet rs = stmt.executeQuery("SELECT eid, entlehnkarte, f_gebuehr(eid) AS gebuehren_add "+
					                         "FROM Entlehnung " +
					                         "WHERE adat = '" + dat + "'");
			
			s = "UPDATE Entlehnkarte SET gebuehr = gebuehr + ? WHERE nr = ?"; 
			
			pstmt = conn.prepareStatement(s);
			
			while (rs.next()) { 
				
				pstmt.setBigDecimal(1, rs.getBigDecimal("gebuehren_add"));
				pstmt.setInt(2, rs.getInt("entlehnkarte"));
				pstmt.executeUpdate();
			}
			
			conn.commit();
			
			/* Statements und ResultSets schliessen */
			
			rs.close();
			stmt.close();
			
		} catch (SQLException e) {
			e.printStackTrace();
			
			if (conn != null) {
				conn.rollback();
			}
			if (stmt != null) {
				stmt.close();
			}
			
		}
	}
	
	/*
	 * Methode "resetGebuehren()", die alle offenen Gebuehren aus der Tabelle "Entlehnkarte" auf 0.00 setzt. 
	 * Die Methode soll auf der Konsole ausgeben, wieviele Datensaetze geaendert wurden.
	 */
	public void resetGebuehren() throws SQLException{
		
		Statement stmt = null;
		int updatedRows = 0;
		
		try {
			stmt = conn.createStatement();
			updatedRows = stmt.executeUpdate("UPDATE Entlehnkarte " +
					                         "SET gebuehr = '0.00' " +
					                         "WHERE gebuehr != '0.00'");
			
			conn.commit();
			stmt.close();
			
		} catch (Exception e) {
			System.out.println(e.toString());
			
		
			if (conn != null) {
				conn.rollback();
			}
			if (stmt != null) {
				stmt.close();
			}
			
		} finally {
			System.out.println("Anzahl aktualisierter Tupel: " + updatedRows);
		}
		
	}
	
	/*
	 * Methode "dbDisconnect()", die die bestehende JDBC-Verbindung wieder schliet.
	 */
	public void dbDisconnect() throws SQLException{
		
			conn.close();
		
	}
	
	/*
	 * Ueberlegen Sie sich eine sinnvolle Testabdeckung der Java-Klasse Bestellungsverwaltung. Legen Sie dazu eine 
	 * weitere Methode testEntlehnungsverwaltung() an, die die anderen Methoden aufruft und entsprechende Ausgaben erzeugt, 
	 * so dass sich die erfolgreiche Durchfuehrung der Tests ueberpruefen laesst. Sie muessen in der Lage sein, diese Tests 
	 * im Rahmen des Abgabegespraechs ablaufen zu lassen.
	 */
	public void testEntlehnungsverwaltung() throws SQLException, ParseException{
		
		String trennlinie = "--------------------------------------------";
		
		/* Zustand vor Aenderungen */
		System.out.println(trennlinie);
		System.out.println("Zustand vor allen Aenderungen");			
		System.out.println(trennlinie);
		dbOutputKunden();
		System.out.println(trennlinie);
		System.out.println("");
			
		/* Gebuehren werden eingetragen */
		System.out.println(trennlinie);
		System.out.println("Gebuehren werden eingetragen.");
		editGebuehren(java.sql.Date.valueOf("2010-10-05"));
		System.out.println("Neuer Gebuehrenstand:");
		System.out.println(trennlinie);
		dbOutputKunden();
		System.out.println(trennlinie);
		System.out.println("");
			
		/* Gebuehren werden zurueckgesetzt */
		System.out.println(trennlinie);
		System.out.println("Gebuehren werden zurueckgesetzt.");
		resetGebuehren();
		System.out.println("Neuer Gebuehrenstand:");
		System.out.println(trennlinie);
		dbOutputKunden();
		System.out.println(trennlinie);
		System.out.println("");
	}
	
	public static void main(String[] args) {
		
		try{
		
			Entlehnungsverwaltung ev = new Entlehnungsverwaltung();
		
			ev.dbConnect();
		
			ev.testEntlehnungsverwaltung();
		
			ev.dbDisconnect();
		}
		catch(SQLException e){
			while(e != null){
				System.out.println("Code: " + e.getErrorCode());
				System.out.println("Message: " + e.getMessage());
				System.out.println("State: " + e.getSQLState());
				e = e.getNextException();
			}
		}
		catch(Exception e){
			e.printStackTrace();
		}
		
	}

}
