/* 5. Trigger:

	Schreiben Sie einen BEFORE INSERT ON - Trigger („t_before_entlehnt“), welcher vor einem Insert 
	in die Tabelle „entlehnt“ ueberprueft, ob ein Exemplar ueberhaupt entlehnt werden darf. Ein 
	Datenbankeintrag soll nur dann stattfinden, wenn...
	
	- der Besitzer der Entlehnkarte die Altersfreigabe nicht verletzt (d.h. wenn ein Medium ab 16 Jahre 
	  freigegeben ist, dann darf es auch nur von mindestens 16 Jahre alten Personen ausgeliehen werden, 
	  nicht aber von juengeren.) Hinweis: zur Berechnung des Alters einer Person koennen Sie die Funktionen 
	  age() und date_part() verwenden (siehe Postgres Dokumentation:
	  http://www.postgresql.org/docs/8.4/static/functions-datetime.html.)
	- der Besitzer der Entlehnkarte die maximal entlehnbare Medienmenge von 10 Exemplaren nicht 
	  ueberschreitet (d.h. pro Entlehnkarte duerfen maximal 10 Exemplare entliehen werden; bereits 
	  zurueckgegebene Exemplare zaehlen selbstverstaendlich nicht dazu!)
	  
	Bei Verletzung von (a), werfen Sie die Exception „zu_jung“ und bei (b) die Exception „zu_viele_exemplare“.*/


CREATE FUNCTION check_entlehnt() RETURNS trigger AS $check_entlehnt$ 
 DECLARE 
  alter_erfuellt INTEGER;
  anz_medien INTEGER;
  
  exemplar_entlehnbar INTEGER;
  
 BEGIN 
  SELECT (CASE WHEN (a.alter >= m.freigabe) THEN '1' ELSE '0' END) AS bedingung INTO alter_erfuellt 
   FROM (SELECT eid, DATE_PART('years', AGE(CURRENT_DATE, p.gebdat)) AS alter 
  FROM Entlehnung e JOIN Entlehnkarte k ON (e.entlehnkarte = k.nr) JOIN Person p ON (k.person = p.pid)) a, 
   Medien m 
  WHERE m.mediennr = NEW.mediennr AND a.eid = NEW.eid;

  IF alter_erfuellt = '0' THEN RAISE EXCEPTION 'zu_jung'; 
  ELSE
   SELECT count(*) INTO anz_medien 
   FROM Entlehnung e NATURAL JOIN entlehnt t 
   WHERE NOT EXISTS (SELECT * FROM retourniert r WHERE r.eid = t.eid AND r.mediennr = t.mediennr AND r.enr = t.enr) 
	AND e.entlehnkarte = (SELECT entlehnkarte FROM entlehnung WHERE eid = NEW.eid)
   GROUP BY e.entlehnkarte;
   
   IF anz_medien >= '10' THEN RAISE EXCEPTION 'zu_viele_exemplare';
   END IF;
  END IF; 
  RETURN NEW; 
END;
$check_entlehnt$ LANGUAGE plpgsql;

CREATE TRIGGER t_before_entlehnt BEFORE INSERT ON entlehnt 
FOR EACH ROW EXECUTE PROCEDURE check_entlehnt();


/* 6.1
   Schreiben Sie eine Funktion ("f_bestseller"), die fuer ein bestimmtes Medium den prozentuellen Anteil 
   der Entlehnungen im Vergleich zur Gesamtanzahl aller Entlehnungen berechnet: 
   - Prozentueller Anteil = (Anzahl der Entlehnungen pro Medium * 100) / (Anzahl der Entlehnungen insgesamt). 
   - Die Funktion hat als Eingabeparameter „mediennummer“ und liefert den prozentuellen Anteil fuer dieses Medium zurueck. 
   - Um bei der Berechnung des prozentuellen Anteils eine Division durch "0" zu vermeiden, stellen Sie bitte sicher, 
     dass ueberhaupt Entlehnungen gespeichert sind. Werfen Sie die Exception "keine_entlehnungen", falls noch ueberhaupt 
	 keine Entlehnungen existieren. 
   - Beachten Sie ausserdem, dass Medien, die noch ueberhaupt nie entlehnt wurden einen prozentuellen Anteil von "0" haben. 
   - Stellen Sie fuer etwaige andere Ausnahmefaelle im Exception-Teil der Funktion "f_bestseller" einen WHEN OTHERS Zweig 
     bereit, der den PostgreSQL-Status (Fehlernummer) und die Fehlermeldung ausgibt.
   - Liefern Sie bei auftretenden Fehlern den Wert "-1" zurueck.*/

CREATE FUNCTION f_bestseller(mediennummer INTEGER) RETURNS INTEGER AS $$
DECLARE
  ist_entlehnt INTEGER;
  anz_medium INTEGER;
  anz_gesamt INTEGER;
  bestseller_wert INTEGER;
BEGIN
  SELECT COUNT(*) INTO anz_gesamt FROM entlehnt;
  IF anz_gesamt = '0' THEN RAISE EXCEPTION 'keine_entlehnungen'; RETURN -1;
  
  ELSE
	SELECT (CASE WHEN (mediennummer NOT IN (SELECT mediennr FROM entlehnt)) THEN '1' ELSE '0' END) 
		AS bedingung INTO ist_entlehnt;
  
	IF ist_entlehnt = '1' THEN
		bestseller_wert := 0;
	
	ELSE 
		SELECT ((COUNT(*)*100)/anz_gesamt) AS prozent INTO bestseller_wert 
		FROM Exemplar NATURAL JOIN entlehnt 
		WHERE mediennr = mediennummer 
		GROUP BY mediennr;
	
	END IF;
  END IF;
  
  RETURN bestseller_wert;
  
EXCEPTION
  WHEN OTHERS THEN RAISE NOTICE 'Fehler Nr. % (%)', SQLSTATE, SQLERRM;
  RETURN -1;
END;
$$ LANGUAGE plpgsql;

  
/* 6.2
   Schreiben Sie eine Funktion („f_gebuehr“), die die anfallenden Gebuehren fuer eine Entlehnung berechnet. 
   Verrechnen Sie fuer jedes Exemplar dieser Entlehnung: 
   ...1.00 Euro, falls es eine Film-DVD ist; 
   ...0.80 Euro, falls es ein Neuerwerb ist - d.h. das Kaufdatum des Exemplars muss weniger als 1 Jahr  
      vor dem aktuellen Datum liegen; 
   ...0.60 Euro, falls der Bestseller-Wert zwischen 20 und 50 Prozent liegt bzw. 1.20 Euro, falls der 
      Bestseller-Wert ueber 50 Prozent liegt. Verwenden Sie die im vorigen Punkt erstellte Funktion 
	  „f_bestseller“, um den Bestseller-Wert zu berechnen. 
   Die Funktion hat als Eingabeparamter „entlehnid“ („eid“ aus Entlehnung) und liefert die Summe der fuer 
   diese Entlehnung anfallenden Gebuehren zurueck. Stellen Sie im Exception-Teil der Funktion "f_gebuehr" 
   einen WHEN OTHERS Zweig bereit, der bei einem unvorhergesehenen Fehler den PostgreSQL-Status (Fehlernummer) 
   und die Fehlermeldung ausgibt. Liefern Sie bei auftretenden Fehlern den Wert "-1" zurueck.*/

CREATE FUNCTION f_gebuehr(entlehnid INTEGER) RETURNS NUMERIC(6,2) AS $$
DECLARE
  anz_filme INTEGER;
  anz_neuerwerb INTEGER;
  anz_bestseller_1 INTEGER;
  anz_bestseller_2 INTEGER;
  gebuehr NUMERIC(6,2);
BEGIN
  SELECT COUNT(*) AS anz INTO anz_filme 
  FROM Entlehnung NATURAL JOIN entlehnt NATURAL JOIN Filme 
  WHERE format = 'DVD' AND eid = entlehnid; 
  
  SELECT COUNT(*) AS anz INTO anz_neuerwerb 
  FROM Entlehnung NATURAL JOIN entlehnt NATURAL JOIN Exemplar 
  WHERE DATE_PART('years', AGE(CURRENT_DATE, kaufdat)) < 1 AND eid = entlehnid;
  
  SELECT COUNT(*) AS anz INTO anz_bestseller_1 
  FROM Entlehnung NATURAL JOIN entlehnt 
  WHERE eid = entlehnid AND f_bestseller(mediennr) <= 50 AND f_bestseller(mediennr) >= 20;
  
  SELECT COUNT(*) AS anz INTO anz_bestseller_2 
  FROM Entlehnung NATURAL JOIN entlehnt 
  WHERE eid = entlehnid AND f_bestseller(mediennr) > 50;
  
  gebuehr := 0.00 + (1.00 * anz_filme) + (0.80 * anz_neuerwerb) + 
    (0.60 * anz_bestseller_1) + (1.20 * anz_bestseller_2);
  RETURN gebuehr;
  
  EXCEPTION
  WHEN OTHERS THEN RAISE NOTICE 'Fehler Nr. % (%)', SQLSTATE, SQLERRM;
  RETURN -1;
END;
$$ LANGUAGE plpgsql;


/* 6.3
   Schreiben Sie eine Prozedur („f_gebuehren_eintragen“), die als Eingabeparameter ein Datum "dat" hat ("adat" von Entlehnung)
   und nichts zurueckgibt (Rueckgabetyp void). Die Prozedur soll fuer alle an diesem Datum stattfindenden Entlehnungen
   die Gebuehren berechenen und auf den Entlehnkarten der jeweiligen Kunden verbuchen ("gebuehr" bei Entlehnkarte). 
   Verwenden Sie zur Berechnung der anfallenden Gebuehren die im vorigen Punkt erstellte Funktion („f_gebuehr“). 
   Beachten Sie dabei, dass sie die „neu“ berechneten Gebuehren zu bereits gespeicherten Gebuehren auf der Entlehnkarte addieren. 
   Stellen Sie im Exception-Teil der Funktion "f_gebuehren_eintragen" einen WHEN OTHERS Zweig bereit, der bei etwaig auftretenden Fehlern 
   den PostgreSQL-Status (Fehlernummer) und die Fehlermeldung ausgibt.*/

CREATE FUNCTION f_gebuehren_eintragen (dat DATE) RETURNS VOID AS $$
DECLARE
  curs refcursor;
  rowvar RECORD;
  
  gebuehren_add NUMERIC(6,2);
  gebuehren_neu NUMERIC(6,2);
  
BEGIN
 
	OPEN curs FOR SELECT e.eid, e.entlehnkarte, k.gebuehr FROM Entlehnung e JOIN Entlehnkarte k ON (e.entlehnkarte = k.nr) WHERE adat = dat;
	FETCH curs INTO rowvar;
	WHILE found LOOP
		gebuehren_add := f_gebuehr(rowvar.eid);
		gebuehren_neu := rowvar.gebuehr + gebuehren_add;
		UPDATE Entlehnkarte SET gebuehr = gebuehren_neu WHERE nr = rowvar.entlehnkarte;
		FETCH curs INTO rowvar;
	END LOOP;
	CLOSE curs;
  
  EXCEPTION
  WHEN OTHERS THEN RAISE NOTICE 'Fehler Nr. % (%)', SQLSTATE, SQLERRM;
END;
$$ LANGUAGE plpgsql;


/* 6.4
   Schreiben Sie eine Prozedur („f_alte_entlehnungen_loeschen()“), die alle Entlehnungen aus der Datenbank 
   entfernt, die laenger als 2 Jahre zurueckliegen, d.h. die Zeitspanne vom Datum der Ausleihe („adat“) und 
   der aktuellen Systemzeit ist gleich bzw. laenger als 2 Jahre. Bedenken Sie, dass Sie auch in der Tabelle "entlehnt" und 
   "retourniert" alle Datensaetze loeschen muessen, die zu den betroffenen Entlehnungen gehoeren. Loeschen Sie 
   allerdings nur jene Entlehnungen, die bereits vollstaendig zurueckgegeben wurden. Die Prozedur soll am Ende 
   einen Hinweis ausgeben, wie viele Entlehnungen entfernt wurden. */

CREATE FUNCTION f_alte_entlehnungen_loeschen() RETURNS VOID AS $$
DECLARE
  curs refcursor;
  rowvar RECORD;
  entfernte_entlehnungen INTEGER;
BEGIN
  entfernte_entlehnungen := 0;

  OPEN curs FOR SELECT e.eid FROM Entlehnung e WHERE DATE_PART('years', AGE(CURRENT_DATE, e.adat)) >= 2 AND 
	EXISTS (SELECT * FROM retourniert WHERE eid = e.eid);
  FETCH curs INTO rowvar;
  WHILE found LOOP
	DELETE FROM retourniert WHERE eid = rowvar.eid;
    DELETE FROM entlehnt WHERE eid = rowvar.eid;
    DELETE FROM Entlehnung WHERE eid = rowvar.eid;
    entfernte_entlehnungen := entfernte_entlehnungen + 1;
    FETCH curs INTO rowvar;
  END LOOP;
  CLOSE curs;

  RAISE NOTICE '% Entlehnungen entfernt.', entfernte_entlehnungen;

EXCEPTION WHEN OTHERS THEN RAISE NOTICE 'Fehler Nr. % (%)', SQLSTATE, SQLERRM;
END;
$$ LANGUAGE plpgsql;
